#include <linux/interrupt.h>
#include <linux/netdevice.h>
#include <linux/rtnetlink.h>
#include <linux/ieee80211.h>
#include <net/mac80211.h>
#include <net/cfg80211.h>
#include "ieee80211_i.h"

#define to_ieee80211_local(class) \
	        container_of(class, struct ieee80211_local, class_dev)

/* For sysfs and debug only */
static struct ieee802_11_elem_tspec _tspec;
static u8 _dls_mac[ETH_ALEN];

#define TSID _tspec.ts_info.tsid
#define TSDIR _tspec.ts_info.direction
#define TSUP _tspec.ts_info.up


/* QoS attributes */
static ssize_t show_ts_info(struct device *dev, struct device_attribute *attr,
			    char *buf)
{
	/* TSID, Direction, UP */
	return sprintf(buf, "%u %u %u\n", TSID, TSDIR, TSUP);
}

static ssize_t store_ts_info(struct device *dev, struct device_attribute *attr,
			     const char *buf, size_t len)
{
	unsigned int id, index, up;

	if (sscanf(buf, "%u, %u, %u", &id, &index, &up) != 3) {
		printk(KERN_ERR "%s: sscanf error\n", __FUNCTION__);
		return -EINVAL;
	}
	if (id < 8 || id > 15) {
		printk(KERN_ERR "invalid tsid %d\n", id);
		return -EINVAL;
	}
	if ((index != 0) && (index != 1) && (index != 3)) {
		printk(KERN_ERR "invalid direction %d\n", index);
		return -EINVAL;
	}
	if (up < 0 || up > 7) {
		printk(KERN_ERR "invalid UP %d\n", up);
		return -EINVAL;
	}
	TSID = id;
	TSDIR = index;
	TSUP = up;
	return len;
}

static DEVICE_ATTR(ts_info, S_IWUSR|S_IRUGO, show_ts_info, store_ts_info);

static ssize_t show_tspec(struct device *dev, struct device_attribute *attr,
			  char *buf)
{
	/* Nominal MSDU, Max MSDU, Min int, Max int, Inact int,
	 * susp int, start, min rate, mean rate, peak rate,
	 * max burst, delay, min phy, surp band, medium time */
	return sprintf(buf,"%hu %hu %u %u %u %u %u %u %u %u %u %u %u %hu %hu\n",
		       _tspec.nominal_msdu_size,
		       _tspec.max_msdu_size,
		       _tspec.min_service_interval,
		       _tspec.max_service_interval,
		       _tspec.inactivity_interval,
		       _tspec.suspension_interval,
		       _tspec.service_start_time,
		       _tspec.min_data_rate,
		       _tspec.mean_data_rate,
		       _tspec.peak_data_rate,
		       _tspec.burst_size,
		       _tspec.delay_bound,
		       _tspec.min_phy_rate,
		       _tspec.surplus_band_allow,
		       _tspec.medium_time);
}

static ssize_t store_tspec(struct device *dev, struct device_attribute *attr,
			   const char *buf, size_t len)
{
	if (sscanf(buf, "%hu %hu %u %u %u %u %u %u %u %u %u %u %u %hu %hu",
		   &_tspec.nominal_msdu_size,
		   &_tspec.max_msdu_size,
		   &_tspec.min_service_interval,
		   &_tspec.max_service_interval,
		   &_tspec.inactivity_interval,
		   &_tspec.suspension_interval,
		   &_tspec.service_start_time,
		   &_tspec.min_data_rate,
		   &_tspec.mean_data_rate,
		   &_tspec.peak_data_rate,
		   &_tspec.burst_size,
		   &_tspec.delay_bound,
		   &_tspec.min_phy_rate,
		   &_tspec.surplus_band_allow,
		   &_tspec.medium_time) != 15) {
		printk(KERN_ERR "%s: sscanf error\n", __FUNCTION__);
		return -EINVAL;
	}
	/* Set default TSPEC Values */
	_tspec.ts_info.access_policy = WLAN_TSINFO_EDCA;
	_tspec.ts_info.apsd = WLAN_TSINFO_PSB_LEGACY;

	return len;
}

static DEVICE_ATTR(tspec, S_IWUSR|S_IRUGO, show_tspec, store_tspec);

static ssize_t store_addts(struct device *dev, struct device_attribute *attr,
			   const char *buf, size_t len)
{
	struct net_device *netdev;

	netdev = to_net_dev(dev);
	ieee80211_send_addts(netdev, &_tspec);
	return len;
}

static DEVICE_ATTR(addts, S_IWUSR, NULL, store_addts);

static ssize_t store_addts_wmm(struct device *dev,
			       struct device_attribute *attr,
			       const char *buf, size_t len)
{
	struct net_device *netdev;

	netdev = to_net_dev(dev);
	wmm_send_addts(netdev, &_tspec);
	return len;
}

static DEVICE_ATTR(addts_wmm, S_IWUSR, NULL, store_addts_wmm);

static ssize_t store_delts(struct device *dev, struct device_attribute *attr,
			   const char *buf, size_t len)
{
	struct net_device *netdev;

	netdev = to_net_dev(dev);
	ieee80211_send_delts(netdev, TSID, TSDIR, _tspec.medium_time);
	return len;
}

static DEVICE_ATTR(delts, S_IWUSR, NULL, store_delts);

static ssize_t store_delts_wmm(struct device *dev,
			       struct device_attribute *attr,
			       const char *buf, size_t len)
{
	struct net_device *netdev;

	netdev = to_net_dev(dev);
	wmm_send_delts(netdev, TSID, TSDIR, _tspec.medium_time);
	return len;
}

static DEVICE_ATTR(delts_wmm, S_IWUSR, NULL, store_delts_wmm);

static ssize_t show_dls_mac(struct device *dev, struct device_attribute *attr,
			    char *buf)
{
	return sprintf(buf, "%02X:%02X:%02X:%02X:%02X:%02X\n",
		       _dls_mac[0], _dls_mac[1], _dls_mac[2],
		       _dls_mac[3], _dls_mac[4], _dls_mac[5]);
}

static ssize_t store_dls_mac(struct device *dev, struct device_attribute *attr,
			     const char *buf, size_t len)
{
	if (sscanf(buf, "%x:%x:%x:%x:%x:%x",
		   (int *)&_dls_mac[0], (int *)&_dls_mac[1],
		   (int *)&_dls_mac[2], (int *)&_dls_mac[3],
		   (int *)&_dls_mac[4], (int *)&_dls_mac[5]) != 6) {
		printk(KERN_ERR "%s: sscanf error\n", __FUNCTION__);
		return -EINVAL;
	}
	return len;
}

static DEVICE_ATTR(dls_mac, S_IWUSR|S_IRUGO, show_dls_mac, store_dls_mac);

static ssize_t show_dls_op(struct device *dev, struct device_attribute *attr,
			   char *buf)
{
	return sprintf(buf, "DLS Operation: Setup = 1; Teardown = 2\n");
}

static ssize_t store_dls_op(struct device *dev, struct device_attribute *attr,
			    const char *buf, size_t len)
{
	struct ieee80211_local *local;
	struct net_device *netdev;
	struct dls_info *dls;
	unsigned int opt;

	netdev = to_net_dev(dev);
	local = wdev_priv(netdev->ieee80211_ptr);

	if (sscanf(buf, "%u", &opt) != 1) {
		printk(KERN_ERR "%s: sscanf error\n", __FUNCTION__);
		return -EINVAL;
	}
	switch (opt) {
	case 1:
		dls = kzalloc(sizeof(struct dls_info), GFP_KERNEL);
		if (!dls) {
			printk(KERN_ERR "No memory to allocate dls_info\n");
			return -ENOMEM;
		}
		atomic_set(&dls->refcnt, 1);
		dls->status = DLS_STATUS_SETUP;
		dls->timeout = 0;
		memcpy(dls->addr, _dls_mac, ETH_ALEN);
		dls_info_add(local, dls);
		ieee80211_send_dls_req(netdev, dls);
		break;
	case 2:
		ieee80211_send_dls_teardown(netdev, _dls_mac,
					    WLAN_REASON_QSTA_NOT_USE);
		break;
	default:
		printk(KERN_ERR "Unknown DLS Operation: %d\n", opt);
		break;
	}
	return len;
}

static DEVICE_ATTR(dls_op, S_IWUSR|S_IRUGO, show_dls_op, store_dls_op);

static void __ieee80211_remove_if_group(struct kobject *kobj,
					struct ieee80211_sub_if_data *sdata)
{
	if (sdata->sysfs_qos_group) {
		sysfs_remove_group(kobj, sdata->sysfs_qos_group);
		sdata->sysfs_qos_group = NULL;
	}
}

static struct attribute *ieee80211_qos_attrs[] = {
	&dev_attr_ts_info.attr,
	&dev_attr_tspec.attr,
	&dev_attr_addts.attr,
	&dev_attr_addts_wmm.attr,
	&dev_attr_delts.attr,
	&dev_attr_delts_wmm.attr,
	&dev_attr_dls_mac.attr,
	&dev_attr_dls_op.attr,
	NULL
};
	
static struct attribute_group ieee80211_qos_group = {
	.name = "qos",
	.attrs = ieee80211_qos_attrs,
};

int ieee80211_sysfs_add_netdevice(struct net_device *dev)
{
	struct kobject *kobj = &dev->dev.kobj;
	struct ieee80211_sub_if_data *sdata = IEEE80211_DEV_TO_SUB_IF(dev);
	int res = 0;

	__ieee80211_remove_if_group(kobj, sdata);
	if (sdata->type == IEEE80211_IF_TYPE_STA) {
		sdata->sysfs_qos_group = &ieee80211_qos_group;
		res = sysfs_create_group(kobj, sdata->sysfs_qos_group);
		if (res)
			sdata->sysfs_qos_group = NULL;
	}
	return res;
}

static inline void ieee80211_remove_if_group(struct kobject *kobj,
                                             struct net_device *dev)
{
	__ieee80211_remove_if_group(kobj, IEEE80211_DEV_TO_SUB_IF(dev));
}

void ieee80211_sysfs_remove_netdevice(struct net_device *dev)
{
	ieee80211_remove_if_group(&dev->dev.kobj, dev);
}




/* general module functions */

static void ieee80211_class_dev_release(struct class_device *dev)
{
	kfree(to_ieee80211_local(dev));
}

static struct class ieee80211_qos_class = {
	.name = "ieee80211_qos",
	.class_dev_attrs = NULL,
	.release = ieee80211_class_dev_release,
};

void ieee80211_dev_sysfs_init(struct ieee80211_local *local)
{
	local->class_dev.class = &ieee80211_qos_class;
	local->class_dev.class_data = local;
	class_device_initialize(&local->class_dev);
}

int ieee80211_sysfs_init(void)
{
	return class_register(&ieee80211_qos_class);
}

void ieee80211_sysfs_deinit(void)
{
	class_unregister(&ieee80211_qos_class);
}
